/*
 ** airdbc_drv.h
 ** AirDBC Database Abstraction Layer Driver Interface.
 **
 ** ------------------------------------------------------------------------
 ** Copyright (C) 2000-2005 Carnegie Mellon University. All Rights Reserved.
 ** ------------------------------------------------------------------------
 ** Authors: Brian Trammell <bht@cert.org>
 ** ------------------------------------------------------------------------
 ** From libair::db, by Brian Trammell, Roman Danyliw, and Sean Levy.
 ** ------------------------------------------------------------------------
 ** GNU Lesser GPL Rights pursuant to Version 2.1, February 1999 
 ** Government Purpose License Rights (GPLR) pursuant to DFARS 252.225-7013
 ** ------------------------------------------------------------------------
 */

#ifndef _ADB_AIRDBC_DRV_H_
#define _ADB_AIRDBC_DRV_H_

/** 
 * @file
 *
 * AirDBC Database Connectivity driver API. This file provides the interface
 * used by AirDBC drivers, and by AirDBC itself to communicate with those
 * drivers. You should not have to include this file unless you are writing
 * a driver for an RDBMS client interface not supported by AirDBC.
 */

struct _AdbDriver;
/** 
 * An AirDBC driver. Generally, an AirDBC driver will create an AdbDriver 
 * structure describing itself, then register that driver with AirDBC via 
 * adb_driver_register(). 
 */
typedef struct _AdbDriver AdbDriver;

/**
 * AirDBC connection internals. May be directly accessed by AirDBC drivers.
 */
struct _AdbConnection {
    /** Driver managing this connection */
    AdbDriver           *driver;
    /** Driver private connection information. Storage for this is managed 
        by the driver. */
    void                *dconn;
    /** AirDBC URI for this connection */
    char                *uri_full;
    /** Storage for URI decomposition */
    char                *uri_store;
    /** URI scheme. Corresponds to the driver's name */
    char                *scheme;
    /** Remote host to connect to */
    char                *host;
    /** Remote port to connect to */
    uint16_t            port;
    /** Username to present to RDBMS on connection */
    char                *username;
    /** Password to present to RDBMS on connection */
    char                *password;
    /** Database name on remote RDBMS server */
    char                *dbname;
    /** Any additional information after the DB name in the URI.
        Available for private use by the driver. */
    char                *additional;
    /** Connection state. Managed by AirDBC; the driver should not modify 
        this field on its own. Connection failures should be cleaned up by
        a call to adb_conn_close. */
    gboolean            is_connected;
};

/**
 * AirDBC statement internals. May be directly accessed by AirDBC drivers.
 */
struct _AdbStatement {
    /** Connection on which this statement was prepared. */
    AdbConnection       *conn;
    /** Driver private statement information. Storage for this is managed
        bu the driver. */
    void                *dstmt;
    /** Statement parameterized string in native parameter format. Generated
         by adb_stmt_prepare() before calling the driver's prepare routine. */
    char                *sql;
    /** Count of parameters in statement. Generated by adb_stmt_prepare() 
        before calling the driver's prepare routine. */
    uint32_t            param_count;
    /** Maximum length of a single parameter in param_buf. This is necessary 
        because AirDBC preallocates parameter space due to the requirements 
        of some RDBMS client libraries. Set by the user at adb_stmt_prepare() 
        time. param_buf is param_maxlen * param_count characters long. */
    uint32_t            param_maxlen;
    /** Preallocated parameter space. Managed by AirDBC. If a position in this 
        buffer is valid (i.e., not NULL), AirDBC will point the appropriate 
        param_str entry into this buffer at adb_stmt_bind() time. */
    char                *param_buf;
    /** Array of parameter strings. These are pointers into param_buf. 
        If NULL for a parameter, a NULL value was bound for that parameter. */
    char                **param_str;
    /** AirDBC-maintained map of parameter names to position lists. */
    GHashTable          *param_map;
};    

/**
 * AirDBC result set internals. May be directly accessed by AirDBC drivers.
 */
struct _AdbResultSet {
    /** Statement from which this result set was generated. */
    AdbStatement        *stmt;
    /** Driver private result set information. Storage for this is managed 
        by the driver. */
    void                *drs;
    /** AirDBC-maintained map of column names to positions. */
    GHashTable          *colname_map;
};

/** Pointer to a driver connection management function */
typedef gboolean    (*AdbConnectionFn)(
    AdbConnection       *conn,
    GError              **err);

/** Pointer to a driver connection free function */
typedef void        (*AdbConnectionFreeFn)(
    AdbConnection       *conn);

/** Pointer to a driver statement management function */
typedef gboolean    (*AdbStatementFn)(
    AdbStatement        *stmt,
    GError              **err);

/** Pointer to a driver statement free function */    
typedef void        (*AdbStatementFreeFn)(
    AdbStatement        *stmt);
    
/** Pointer to a driver execute function. */
typedef gboolean    (*AdbExecuteFn)(
    AdbStatement        *stmt,
    GError              **err);

/** Pointer to a driver query function. */
typedef gboolean    (*AdbQueryFn)(
    AdbResultSet        *rs,
    GError              **err);

/** Pointer to a driver result set free function */        
typedef void        (*AdbResultSetFreeFn)(
    AdbResultSet        *rs);

/** Pointer to a driver result set management function */        
typedef gboolean    (*AdbResultSetFn)(
    AdbResultSet        *rs,
    GError              **err);

/** Pointer to a driver result set column count function */        
typedef uint32_t    (*AdbResultSetCountFn)(
    AdbResultSet        *rs,
    GError              **err);

/** Pointer to a driver result set field width function */
typedef uint32_t (*AdbResultSetFieldWidthFn)(
    AdbResultSet *rs,
    uint32_t      row,
    uint32_t      col,
    GError      **err);

/** Pointer to a driver result set column name function */        
typedef char        *(*AdbResultSetNameFn)(
    AdbResultSet        *rs,
    uint32_t            col,
    GError              **err);

/** Pointer to a driver result set value fetch function */
typedef gboolean (*AdbFetchFn)(
    AdbResultSet         *rs,
    uint32_t              col,
    const char          **val,
    GError              **err);

/** Pointer to a driver result set value fetch into buffer function */
typedef gboolean (*AdbFetchBufFn)(
    AdbResultSet         *rs,
    uint32_t              col,
    const char           *buf,
    size_t                len,
    GError              **err);

/**
 * AirDBC driver internals. May be directly accessed by AirDBC drivers.
 * Contains a set of function pointers that define the driver's actions.
 */
struct _AdbDriver {
    /** Driver name. Corresponds to the URI scheme which will select 
        this driver. */
    const char          *name;
    /** Native SQL parameter prefix string. Each named prefix in the 
        expression passed to adb_stmt_prepare() will be replaced with a 
        string beginning with this substring. */
    const char          *param_prefix;
    /** TRUE if native SQL positional parameters are numbered. */
    gboolean            param_numsuffix;
    /** Driver private allocation function. Called by adb_conn_create() to 
        allow driver to set up driver private information for the connection. */
    AdbConnectionFn     alloc;
    /** Driver connection open function. Called by adb_conn_open() to open
        the connection to the backend. */
    AdbConnectionFn     open;
    /** Driver connection open function. Called by adb_conn_close() to close 
        the connection to the backend. */
    AdbConnectionFn     close;
    /** Driver private disposal function. Called by adb_conn_free() to free
        driver private information. */
    AdbConnectionFreeFn free;
    /** Driver transaction begin function. */
    AdbConnectionFn     begin;
    /** Driver transaction commit function. */
    AdbConnectionFn     commit;
    /** Driver transaction rollback function. */
    AdbConnectionFn     rollback;
    /** Driver statement prepare function. Called by adb_stmt_prepare() to 
        allow driver to prepare native statement and set up driver private 
        statement information for the statement. */
    AdbStatementFn      prepare;
    /** Driver private statement disposal function. Called by adb_stmt_free() 
        to free driver private statement information. */
    AdbStatementFreeFn  free_stmt;
    /** Driver statement execution function. */
    AdbExecuteFn        exec;
    /** Driver statement query function. Called by adb_stmt_query() with a 
        pre-allocated AdbResultSet to allow the driver to execute the query 
        and set up driver private result set information. */
    AdbQueryFn          query;
    /** Driver private result set disposal function. Called by adb_rs_free() 
        to free driver private result set information. */
    AdbResultSetFreeFn  free_rs;
    /** Driver result set advance function. Your driver must be set up to require adb_rs_next() to be called before first fetch (i.e., a next call before a fetch should not skip the first row. */
    AdbResultSetFn      next;
    /** Driver result set row count function. */
    AdbResultSetCountFn rowcount;
    /** Driver result set column count function. */
    AdbResultSetCountFn colcount;
    /** Driver result set field width function. */
    AdbResultSetFieldWidthFn fieldwidth;
    /** Driver result set column name function. */
    AdbResultSetNameFn  colname;
    /** Driver result set fetch function. May return pointers to 
        driver private information; the AirDBC API specifies that 
        applications should copy data returned from fetch if it must 
        be accessed beyond a call to adb_rs_next(). */
    AdbFetchFn          fetch;
    /** Driver result set fetch into buffer function.  The caller is responsible
    for managing the storage for the input buffer, and this function will copy
    the fetched result into this buffer. */
    AdbFetchBufFn          fetch_buf;
};

/** Register a driver with AirDBC. After this call, all adb_conn_create() 
    calls with a URI scheme referencing this driver invoke this driver. Since
    adb_conn_create() automatically pre-registers builtin drivers, you should
    only call this for your own drivers. You should provide an API call in 
    your driver to allow applications to register your driver before attempting
    to connect using it. See the built-in driver source for examples.
    
    @param driver pointer to driver definition structure to register. AirDBC 
                  will not copy this structure, so it is unsafe to free any 
                  driver structure storage after registering it. Drivers 
                  should be statically allocated for this reason.
*/
void                adb_driver_register(
    AdbDriver           *driver);

#endif
